# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

from .storage import Array, AOS, SOA

class BaseContainer:
    gpu_support = False
    def __init__(self, domain, scheme, sorder, default_type):
        self.dim = domain.dim
        self.mpi_topo = domain.mpi_topo

        self.nv = scheme.stencil.nv_ptr[-1]
        self.nspace = domain.global_size
        self.vmax = domain.stencil.vmax
        self.sorder = sorder

        if sorder:
            self.m = Array(self.nv, self.nspace, self.vmax, sorder, self.mpi_topo, gpu_support=self.gpu_support)
            self.F = Array(self.nv, self.nspace, self.vmax, sorder, self.mpi_topo, gpu_support=self.gpu_support)
        else:
            self.m = default_type(self.nv, self.nspace, self.vmax, self.mpi_topo, gpu_support=self.gpu_support)
            self.F = default_type(self.nv, self.nspace, self.vmax, self.mpi_topo, gpu_support=self.gpu_support)
            sorder = [i for i in range(self.dim + 1)]

        self.m.set_conserved_moments(scheme.consm)
        self.F.set_conserved_moments(scheme.consm)

        self._set_sorder(sorder)

    def _set_sorder(self, sorder):
        pass

class NumpyContainer(BaseContainer):
    def __init__(self, domain, scheme, sorder=None, default_type=SOA):
        super(NumpyContainer, self).__init__(domain, scheme, sorder, default_type)
        self.Fnew = self.F

    def _set_sorder(self, sorder):
        if not self.sorder:
            self.sorder = [i for i in range(self.dim + 1)]

class CythonContainer(BaseContainer):
    def __init__(self, domain, scheme, sorder=None, default_type=AOS):
        super(CythonContainer, self).__init__(domain, scheme, sorder, default_type)
        self.Fnew = Array(self.nv, self.nspace, self.vmax, self.sorder, self.mpi_topo, gpu_support=self.gpu_support)
        self.Fnew.set_conserved_moments(scheme.consm)

    def _set_sorder(self, sorder):
        if not self.sorder:
            self.sorder = [self.dim] + [i for i in range(self.dim)]

class LoopyContainer(CythonContainer):
    gpu_support = True
    def __init__(self, domain, scheme, sorder=None, default_type=AOS):
        super(LoopyContainer, self).__init__(domain, scheme, sorder, default_type)

    def move2gpu(self, array):
        try:
            import pyopencl as cl
            import pyopencl.array #pylint: disable=unused-variable
            from .context import queue
        except ImportError:
            raise ImportError("Please install loo.py")
        return cl.array.to_device(queue, array)


# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Geometry module
"""

import logging
# from textwrap import dedent
# from six import string_types
import numpy as np

from . import viewer
from .validator import validate

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def get_box(dico):
    """
    return the dimension and the bounds of the box defined in the dictionnary.

    Parameters
    ----------

    dico : dict

    Returns
    -------

    int
        the dimension of the box

    ndarray
        the bounds of the box

    """
    try:
        box = dico['box']
        try:
            bounds = [box['x']]
            dim = 1
            boxy = box.get('y', None)
            if boxy is not None:
                bounds.append(boxy)
                dim += 1
                boxz = box.get('z', None)
                if boxz is not None:
                    bounds.append(boxz)
                    dim += 1
        except KeyError:
            err_msg = "'x' interval not found "
            err_msg += "in the box definition of the geometry."
            log.error(err_msg)
    except KeyError:
        err_msg = "'box' key not found "
        err_msg += "in the geometry definition. "
        err_msg += "Check the input dictionnary."
        log.error(err_msg)
    return dim, np.asarray(bounds, dtype='f8')


class Geometry:
    """
    Create a geometry that defines the fluid part and the solid part.

    Parameters
    ----------

    dico : dict
        dictionary that contains the following `key:value`
            - box : a dictionary for the definition of the computed box
            - elements : a list of elements (optional)
    need_validation : bool
        boolean to specify if the dictionary has to be validated (optional)

    Attributes
    ----------

    dim : int
        number of spatial dimensions (1, 2, or 3)
    bounds : ndarray
        the bounds of the box in each spatial direction
    box_label : list
        a list of the four labels for the
            left, right, bottom, top, front, and back edges
    list_elem : list
        a list that contains each element added or deleted in the box

    Notes
    -----

    The dictionary that defines the box should contains the following
    `key:value`

        - x : a list of the bounds in the first direction
        - y : a list of the bounds in the second direction (optional)
        - z : a list of the bounds in the third direction (optional)
        - label : an integer or a list of integers
            (length twice the number of dimensions)
            used to label each edge (optional)

    Examples
    --------

    see demo/examples/geometry/
    """

    def __init__(self, dico, need_validation=True):
        if need_validation:
            validate(dico, __class__.__name__)

        self.dim, self.bounds = get_box(dico)

        self.list_elem = []

        dummylab = dico['box'].get('label', -1)
        if isinstance(dummylab, int):
            self.box_label = [dummylab]*2*self.dim
        elif isinstance(dummylab, list):
            if len(dummylab) != 2*self.dim:
                err_msg = "The list label of the box has the wrong size "
                err_msg += "(must be 2*dim)"
                log.error(err_msg)
            self.box_label = dummylab
        else:
            log.error("The labels of the box must be an integer or a list")

        self.bounds = np.asarray(self.bounds, dtype='f8')

        log.debug("Message from geometry.py (box_label):\n %d", self.box_label)
        log.debug("Message from geometry.py (bounds):\n %s", self.bounds)

        elem = dico.get('elements', None)
        if elem is not None:
            for elemk in elem:
                if elemk.dim != self.dim:
                    raise ValueError(
                        "Element must have the same dimension of the box"
                    )
                self.list_elem.append(elemk)
        log.debug(self.__str__())

    def __str__(self):
        from .utils import header_string
        from .jinja_env import env
        template = env.get_template('geometry.tpl')
        return template.render(
            header=header_string('Geometry information'),
            geom=self
        )

    def __repr__(self):
        return self.__str__()

    def add_elem(self, elem):
        """
        add a solid or a fluid part in the domain

        Parameters
        ----------

        elem : Element
            a geometric element to add (or to del)

        """
        self.list_elem.append(elem)

    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    # pylint: disable=too-complex
    def visualize(self,
                  viewer_app=viewer.matplotlib_viewer,
                  figsize=(6, 4),
                  viewlabel=False,
                  fluid_color='navy',
                  viewgrid=False,
                  alpha=1.):
        """
        plot a view of the geometry

        Parameters
        ----------

        viewer_app : Viewer
            a viewer (default matplotlib_viewer)
        viewlabel : boolean
            activate the labels mark (default False)
        fluid_color : color
            color for the fluid part (default blue)
        figsize : tuple
            the size of the figure (default (6, 4))
        viewgrid : bool
            view the grid (default False)
        alpha : double
            transparency between 0 and 1 (default 1)


        Returns
        -------

        object
            views

        """
        views = viewer_app.Fig(dim=self.dim, figsize=figsize)
        view = views[0]

        if self.dim == 1:
            xmin, xmax = self.bounds[0][:]
            dist = xmax - xmin
            height = dist/20
            length = dist/50
            lpos = np.asarray([[xmin + length, xmin, xmin, xmin + length],
                               [-height, -height, height, height]]).T
            pos = np.asarray([[xmin, xmax], [0, 0]]).T
            rpos = np.asarray([[xmax - length, xmax, xmax, xmax - length],
                               [-height, -height, height, height]]).T
            view.line(lpos, color=fluid_color)
            view.line(rpos, color=fluid_color)
            view.line(pos, color=fluid_color)
            if viewlabel:
                # label 0 for left
                view.text(str(self.box_label[0]), [xmin + length, -2*height])
                # label 1 for right
                view.text(str(self.box_label[1]), [xmax - length, -2*height])
            view.axis(xmin - dist/2, xmax + dist/2, -10*height, 10*height)
            view.yaxis_set_visible(False)
            view.xaxis_set_visible(viewgrid)
        elif self.dim == 2:
            xmin, xmax = self.bounds[0][:]
            ymin, ymax = self.bounds[1][:]
            view.polygon(np.array([[xmin, ymin],
                                   [xmin, ymax],
                                   [xmax, ymax],
                                   [xmax, ymin]]), fluid_color, alpha=alpha)
            if viewlabel:
                # label 0 for left
                view.text(str(self.box_label[0]), [xmin, 0.5*(ymin+ymax)])
                # label 1 for right
                view.text(str(self.box_label[1]), [xmax, 0.5*(ymin+ymax)])
                # label 2 for bottom
                view.text(str(self.box_label[2]), [0.5*(xmin+xmax), ymin])
                # label 3 for top
                view.text(str(self.box_label[3]), [0.5*(xmin+xmax), ymax])
            for elem in self.list_elem:
                if elem.isfluid:
                    color = fluid_color
                    alpha_ = alpha
                else:
                    color = 'white'
                    alpha_ = 1
                elem.visualize(view, color, viewlabel, alpha=alpha_)
            xpercent = 0.05*(xmax-xmin)
            ypercent = 0.05*(ymax-ymin)
            view.axis(
                xmin-xpercent,
                xmax+xpercent,
                ymin-ypercent,
                ymax+ypercent,
                aspect='equal'
            )
            view.grid(viewgrid)
        elif self.dim == 3:
            couleurs = [(.5+.5/k, .5/k, 1.-1./k) for k in range(1, 11)]
            pmin = [(float)(self.bounds[k][0]) for k in range(3)]
            pmax = [(float)(self.bounds[k][1]) for k in range(3)]
            xmin, xmax = pmin[0], pmax[0]
            ymin, ymax = pmin[1], pmax[1]
            zmin, zmax = pmin[2], pmax[2]
            ct_lab = 0
            for k in range(3):
                for x0_ in [pmin[k], pmax[k]]:
                    xgrid, ygrid = np.meshgrid(
                        [pmin[(k+1) % 3], pmax[(k+1) % 3]],
                        [pmin[(k+2) % 3], pmax[(k+2) % 3]]
                    )
                    zgrid = x0_ + np.zeros(xgrid.shape)
                    coord = [xgrid, ygrid, zgrid]
                    view.surface(
                        coord[(2-k) % 3],
                        coord[(3-k) % 3],
                        coord[(1-k) % 3],
                        color=couleurs[self.box_label[ct_lab] % 10],
                        alpha=min(alpha, 0.5)
                    )
                    if viewlabel:
                        x = .25*np.sum(coord[(2-k) % 3])
                        y = .25*np.sum(coord[(3-k) % 3])
                        z = .25*np.sum(coord[(1-k) % 3])
                        view.text(
                            str(self.box_label[ct_lab]),
                            [x, y, z],
                            fontsize=18
                        )
                    ct_lab += 1
            view.axis(xmin, xmax, ymin, ymax, zmin, zmax, aspect='equal')
            view.set_label("X", "Y", "Z")
            for elem in self.list_elem:
                if elem.isfluid:
                    color = fluid_color
                    alpha_ = alpha
                else:
                    color = [
                        couleurs[elem.label[k]]
                        for k in range(elem.number_of_bounds)
                    ]
                    alpha_ = 1
                elem.visualize(view, color, viewlabel, alpha=alpha_)
        else:
            err_msg = "Error in geometry.visualize(): "
            err_msg += "the dimension {:d} is not allowed".format(self.dim)
            log.error(err_msg)

        views.title = "Geometry"
        views.show()
        return views

    def list_of_labels(self):
        """
        Get the list of all the labels used in the geometry.
        """
        labels = np.unique(self.box_label)
        return np.union1d(labels, self.list_of_elements_labels())

    def list_of_elements_labels(self):
        """
        Get the list of all the labels used in the geometry.
        """
        labels = np.empty(0)
        for elem in self.list_elem:
            labels = np.union1d(labels, elem.label)
        return labels



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

# pylint: disable=invalid-name, missing-docstring

"""
utils module
"""

import sys
from colorama import Fore, Style, Back  # pylint: disable=unused-import


def header_string(title):
    barre = '+' + '-'*(len(title)+2) + '+'
    output = '\n| %s |\n' % title
    return Fore.BLUE + barre + output + barre + Fore.RESET


class itemproperty(property):
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        else:
            return bounditemproperty(self, obj)


class item2property(property):
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        else:
            return bounditemproperty(self, obj, True)


class bounditemproperty:
    def __init__(self, item_property, instance, nextItem=False):
        self.__item_property = item_property
        self.__instance = instance
        self.nextItem = nextItem

    def __getitem__(self, key):
        fget = self.__item_property.fget
        if fget is None:
            raise AttributeError("unreadable attribute item")
        if self.nextItem:
            return bound2itemproperty(
                self.__item_property,
                self.__instance,
                key
            )
        else:
            return fget(self.__instance, key)

    def __setitem__(self, key, value):
        fset = self.__item_property.fset
        if fset is None:
            raise AttributeError("can't set attribute item")
        fset(self.__instance, key, value)


class bound2itemproperty:
    def __init__(self, item_property, instance, key):
        self.__item_property = item_property
        self.__instance = instance
        self.key = key

    def __getitem__(self, key):
        fget = self.__item_property.fget
        if fget is None:
            raise AttributeError("unreadable attribute item")
        return fget(self.__instance, self.key, key)

    def __setitem__(self, key, value):
        fset = self.__item_property.fset
        if fset is None:
            raise AttributeError("can't set attribute item")
        fset(self.__instance, self.key, key, value)


def hsl_to_rgb(h, s, l):
    """
    Converts an HSL color value to RGB. Conversion formula
    adapted from http://en.wikipedia.org/wiki/HSL_color_space.
    Assumes h, s, and l are contained in the set [0, 1] and
    the output r, g, and b are in the set [0, 1].

    Parameters
    ----------

    h : double
        the hue
    s : double
        the saturation
    l : double
        the lightness

    Returns
    -------

    tuple
        the color in RGB format
    """

    if s == 0:  # achromatic
        r = l
        g = l
        b = l
    else:
        def hue2rgb(p, q, t):
            t = t % 1
            if t < 1/6:
                return p + (q - p) * 6 * t
            if t < 1/2:
                return q
            if t < 2/3:
                return p + (q - p) * (2/3 - t) * 6
            return p

        q = l * (1 + s) if l < 0.5 else l + s - l * s
        p = 2 * l - q
        r = hue2rgb(p, q, h + 1/3)
        g = hue2rgb(p, q, h)
        b = hue2rgb(p, q, h - 1/3)

    return r, g, b


def print_progress(iteration, total,
                   prefix='', suffix='',
                   decimals=1, barLength=100
                   ):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals
                                  in percent complete (Int)
        barLength   - Optional  : character length of bar (Int)
    """
    formatStr = "{0:." + str(decimals) + "f}"
    percents = formatStr.format(100 * (iteration / float(total)))
    filledLength = int(round(barLength * iteration / float(total)))
    p_bar = '*' * filledLength + '-' * (barLength - filledLength)
    # pylint: disable=expression-not-assigned
    sys.stdout.write(
        '\r%s |%s| %s%s %s' % (prefix, p_bar, percents, '%', suffix)
    ),
    sys.stdout.flush()
    if iteration == total:
        sys.stdout.write('\n')
        sys.stdout.flush()



# FIXME: make pylint happy !
#pylint: disable=all

from sympy.utilities.codegen import CodeGen, CodeGenError, ResultBase, Result, InputArgument, InOutArgument, OutputArgument
from sympy.core import Symbol, S, Expr, Tuple, Equality, Function, sympify
from sympy.core.compatibility import is_sequence, StringIO, string_types
from sympy.printing.codeprinter import AssignmentError
from sympy.core.sympify import _sympify, sympify

from sympy.tensor import Idx, Indexed, IndexedBase
from sympy.matrices import (MatrixSymbol, ImmutableMatrix, MatrixBase,
                            MatrixExpr, MatrixSlice)
from sympy.core.basic import Basic

default_settings = {'export': True}

from .ast import For, If
from .printing.cython import cython_code, CythonCodePrinter
from .printing.numpy import numpy_code, NumpyCodePrinter
from .printing.loopy import loopy_code, LoopyCodePrinter

class Routine(object):
    """Generic description of evaluation routine for set of expressions.

    A CodeGen class can translate instances of this class into code in a
    particular language.  The routine specification covers all the features
    present in these languages.  The CodeGen part must raise an exception
    when certain features are not present in the target language.  For
    example, multiple return values are possible in Python, but not in C or
    Fortran.  Another example: Fortran and Python support complex numbers,
    while C does not.

    """

    def __init__(self, name, arguments, instructions, idx_vars, local_vars, settings={}):
        """Initialize a Routine instance.

        Parameters
        ==========

        name : string
            Name of the routine.

        arguments : list of Arguments
            These are things that appear in arguments of a routine, often
            appearing on the right-hand side of a function call.  These are
            commonly InputArguments but in some languages, they can also be
            OutputArguments or InOutArguments (e.g., pass-by-reference in C
            code).

        instructions : list
            Instructions of the routine.

        local_vars : list of Symbols
            These are used internally by the routine.

        global_vars : list of Symbols
            Variables which will not be passed into the function.

        """

        # extract all input symbols and all symbols appearing in an expression
        input_symbols = set([])
        symbols = set([])
        for arg in arguments:
            if isinstance(arg, OutputArgument):
                symbols.update(arg.expr.free_symbols)
            elif isinstance(arg, InputArgument):
                input_symbols.add(arg.name)
            elif isinstance(arg, InOutArgument):
                input_symbols.add(arg.name)
                symbols.update(arg.expr.free_symbols)
            else:
                raise ValueError("Unknown Routine argument: %s" % arg)

        for i in instructions:
            symbols.update(i.free_symbols)

        symbols = set([s.label if isinstance(s, Idx) else s for s in symbols])

        # Check that all symbols in the expressions are covered by
        # InputArguments/InOutArguments---subset because user could
        # specify additional (unused) InputArguments or local_vars.
        dummy = [i.label for i in idx_vars]
        symbol_indexed_local = set()
        for l in local_vars:
            for ll in l.atoms(IndexedBase):
                symbol_indexed_local.update(ll.atoms(Symbol) - ll.shape.atoms(Symbol))

        notcovered = symbols.difference(
            input_symbols.union(dummy).union(local_vars).union(symbol_indexed_local))
        if notcovered != set([]):
            raise ValueError("Symbols needed for output are not in input " +
                             ", ".join([str(x) for x in notcovered]))

        self.name = name
        self.arguments = arguments
        self.instructions = instructions
        self.idx_vars = idx_vars
        self.local_vars = local_vars
        self.settings = settings

    def __str__(self):
        return self.__class__.__name__ + "({name!r}, {arguments}, {instructions}, {idx_vars}, {local_vars})".format(**self.__dict__)

    __repr__ = __str__

    @property
    def variables(self):
        """Returns a set of all variables possibly used in the routine.

        For routines with unnamed return values, the dummies that may or
        may not be used will be included in the set.

        """
        v = set(self.local_vars)
        for arg in self.arguments:
            v.add(arg.name)
        for res in self.results:
            v.add(res.result_var)
        return v


def get_dims_and_symbol(expr):
    if isinstance(expr, Indexed):
        dims = tuple([ (S.Zero, dim - 1) for dim in expr.shape])
        symbol = expr.base.label
    elif isinstance(expr, Symbol):
        dims = []
        symbol = expr
    elif isinstance(expr, MatrixSymbol):
        dims = tuple([ (S.Zero, dim - 1) for dim in expr.shape if dim != 1])
        symbol = expr
    elif isinstance(expr, MatrixBase):
        # if we have a Matrix, we set line by line the code
        # useful when you have indexes like
        # A[i, j, 0] = B[i, j, 4]

        # todo: regarder que les matrices pour le in et out ont la meme taille
        # todo: regarder si le symbole est toujours le meme dans le terme de gauche (expr)
        for i in range(expr.shape[0]):
            symbol = expr[i].base.label
            dims = tuple([ (S.Zero, dim - 1) for dim in expr[i].base.shape if dim != 1])
    elif isinstance(expr, MatrixSlice):
        symbol = expr.parent
        dims = tuple([ (S.Zero, dim - 1) for dim in symbol.shape if dim != 1])
    else:
        raise CodeGenError("Only Indexed, Symbol, or MatrixSymbol "
                            "can define output arguments.")
    return dims, symbol

def extract(expressions, symbols):
    # extract arguments and instructions of the routine
    output_args = []
    instructions = []
    for expr in expressions:
        instructions.append(expr)
        if isinstance(expr, Equality):
            out_arg = expr.lhs
            expr = expr.rhs
            dims, symbol = get_dims_and_symbol(out_arg)

            if symbol in symbols:
                output_args.append(
                    InOutArgument(symbol, out_arg, expr, dimensions=dims))

                # avoid duplicate arguments
                symbols.remove(symbol)
        elif isinstance(expr, For):
            args, vals = extract(expr.expr, symbols)
            output_args += args
        elif isinstance(expr, If):
            for c, e in expr.statement:
                args, vals = extract(e, symbols)
                output_args += args
        else:
            raise TypeError("The expression must be a For or an equality (Eq).")
    return output_args, instructions


class LBMCodeGen(CodeGen):
    """Generator for Cython code.

    The .write() method inherited from CodeGen will output a code file <prefix>.pyx.

    """

    code_extension = None

    def routine(self, name, expr, argument_sequence, local_vars, settings):
        """Specialized Routine creation for Cython."""

        if is_sequence(expr) and not isinstance(expr, (MatrixBase, MatrixExpr)):
            if not expr:
                raise ValueError("No expression given")
            expressions = Tuple(*expr)
        else:
            expressions = Tuple(expr)

        self.settings = settings

        # local variables
        idx_vars = set()
        symbol_idx_vars = set()
        for l in expressions.atoms(For):
            idx_vars.update({i for i in l.atoms(Idx)})
            # remove symbols that have the same name of Idx in loop
            name_idx = [i.label.name for i in l.atoms(Idx)]
            symbol_idx_vars.update({i for i in l.atoms(Symbol) if i.name in name_idx})

        score_table = {}
        for i in idx_vars:
            score_table[i] = 0

        def rate_index_position(p):
            return p*5

        arrays = expressions.atoms(Indexed)
        for arr in arrays:
            for p, ind in enumerate(arr.indices):
                try:
                    score_table[ind] += rate_index_position(p)
                except KeyError:
                    pass

        idx_order = sorted(idx_vars, key=lambda x: score_table[x])

        # local variables
        local_vars = set() if local_vars is None else set(local_vars)

        # symbols that should be arguments
        symbol_indexed_local = set()
        for l in local_vars:
            for ll in l.atoms(IndexedBase):
                symbol_indexed_local.update(ll.atoms(Symbol) - ll.shape.atoms(Symbol))

        symbols = expressions.free_symbols - idx_vars - local_vars - symbol_idx_vars - symbol_indexed_local

        new_symbols = set([])
        new_symbols.update(symbols)

        for symbol in symbols:
            if isinstance(symbol, Idx):
                new_symbols.remove(symbol)
                if symbol.label in idx_vars:
                    new_symbols.update(symbol.args[1].free_symbols)
                else:
                    new_symbols.update([symbol.label])
        symbols = new_symbols

        output_args, instructions = extract(expressions, symbols)

        arg_list = []

        # setup input argument list
        array_symbols = {}
        for array in expressions.atoms(Indexed):
            array_symbols[array.base.label] = array
        for array in expressions.atoms(MatrixSymbol):
            array_symbols[array] = array

        for symbol in sorted(symbols, key=str):
            if symbol in array_symbols:
                dims = []
                array = array_symbols[symbol]
                for dim in array.shape:
                    if dim != 1:
                        dims.append((S.Zero, dim - 1))
                metadata = {'dimensions': dims}
            else:
                metadata = {}

            arg_list.append(InputArgument(symbol, **metadata))

        output_args.sort(key=lambda x: str(x.name))
        arg_list.extend(output_args)

        if argument_sequence is not None:
            # if the user has supplied IndexedBase instances, we'll accept that
            new_sequence = []
            for arg in argument_sequence:
                if isinstance(arg, IndexedBase):
                    new_sequence.append(arg.label)
                else:
                    new_sequence.append(arg)
            argument_sequence = new_sequence

            missing = [x for x in arg_list if x.name not in argument_sequence]
            if missing:
                msg = "Argument list didn't specify: {0} "
                msg = msg.format(", ".join([str(m.name) for m in missing]))
                raise CodeGenArgumentListError(msg, missing)

            # create redundant arguments to produce the requested sequence
            name_arg_dict = {x.name: x for x in arg_list}
            new_args = []
            for symbol in argument_sequence:
                try:
                    new_args.append(name_arg_dict[symbol])
                except KeyError:
                    new_args.append(InputArgument(symbol))
            arg_list = new_args

        return Routine(name, arg_list, instructions, idx_order, local_vars, settings)

    def code_generator(self, expr, assign_to=None):
        pass

    def _get_symbol(self, s):
        """Print the symbol appropriately."""
        return self.code_generator(s).strip()

    def _call_printer(self, routine):
        code_lines = []

        # Compose a list of symbols to be dereferenced in the function
        # body. These are the arguments that were passed by a reference
        # pointer, excluding arrays.
        dereference = []
        for arg in routine.arguments:
            if isinstance(arg, ResultBase) and not arg.dimensions:
                dereference.append(arg.name)

        for instruction in routine.instructions:
            constants, not_supported, expr = self.code_generator(instruction, human=False)
            code_lines.append("%s\n" % (expr))
        return code_lines

        # declarations = []
        # code_lines = []
        # for i, result in enumerate(routine.results):
        #     print(self.settings)
        #     constants, not_supported, jl_expr = self.code_generator(result, human=False, **self.settings)

        #     for obj, v in sorted(constants, key=str):
        #         declarations.append(
        #             "%s = %s\n" % (obj, v))
        #     for obj in sorted(not_supported, key=str):
        #         if isinstance(obj, Function):
        #             name = obj.func
        #         else:
        #             name = obj
        #         declarations.append(
        #             "# unsupported: %s\n" % (name))
        #     code_lines.append("%s\n" % (jl_expr))
        # return declarations + code_lines

    def _get_routine_opening(self, routine):
        return []

    def _get_routine_ending(self, routine):
        return []


class CythonCodeGen(LBMCodeGen):
    """Generator for Cython code.

    The .write() method inherited from CodeGen will output a code file <prefix>.pyx.

    """

    code_extension = "pyx"

    def code_generator(self, expr, assign_to=None, **settings):
        return cython_code(expr, assign_to, **settings)

    def _get_header(self):
        code_lines = ["#!python\n",
                      "#cython: boundscheck=False\n",
                      "#cython: wraparound=False\n",
                      "#cython: cdivision=True\n",
                      "#cython: binding=True\n",
                      "#import cython\n",
                      "from libc.math cimport *\n",
                      "import numpy as np\n",
                     ]
        return code_lines + ["\n\n"]

    def _preprocessor_statements(self, prefix):
        # code_lines = ["#!python\n",
        #               "#cython: boundscheck=False\n",
        #               "#cython: wraparound=False\n",
        #               "#cython: cdivision=True\n",
        #               "#cython: binding=True\n",
        #               "#import cython\n",
        #               "from libc.math cimport *\n",
        #              ]
        # return code_lines + ["\n\n"]
        return []

    def _get_routine_opening(self, routine):
        """Returns the opening statements of the routine."""
        code_list = []
        export = True
        # export = self.settings.pop('export', True)
        if export:
            code_list.append("def ")
        else:
            code_list.append("cdef void ")

        # Inputs
        args = []
        for i, arg in enumerate(routine.arguments):
            if isinstance(arg, OutputArgument):
                raise CodeGenError("Cython: invalid argument of type %s" %
                                   str(type(arg)))

            if isinstance(arg, (InputArgument, InOutArgument)):
                name = self._get_symbol(arg.name)
                if not arg.dimensions:
                    # If it is a scalar
                    if isinstance(arg, ResultBase):
                        # if it is an output
                        args.append((arg.get_datatype('C'), "*%s" % name))
                    else:
                        # if it is an output
                        args.append((arg.get_datatype('C'), name))
                else:
                    if not export and len(arg.dimensions) == 1:
                        # if the dimension is 1
                        args.append((arg.get_datatype('C'), "*%s" % name))
                    else:
                        args.append((arg.get_datatype('C') + '[' + ', '.join([':']*len(arg.dimensions)) + ':1]', "%s" % name))

        args = ", ".join([ "%s %s" % t for t in args])
        code_list.append("%s(%s)%s\n" % (routine.name, args, ":" if export else " nogil:"))
        code_list = [ "".join(code_list) ]

        return code_list

    def _declare_arguments(self, routine):
        return []

    def _declare_globals(self, routine):
        args = []
        for g in routine.local_vars:
            if isinstance(g, Symbol):
                args.append("cdef double %s\n"%(self._get_symbol(g)))
            else:
                shape = [d for d in g.shape if d!=1]
                if isinstance(g, Indexed):
                    args.append("cdef double[%s] %s = np.zeros((%s))\n"%(', '.join([':']*len(shape)) + ':1', g.base, ','.join("%s"%s for s in shape)))
                else:
                    args.append("cdef double %s[%s]\n"%(self._get_symbol(g), ','.join("%s"%s for s in shape)))
        return ["".join(args)]

    def _declare_locals(self, routine):
        s = []
        for l in routine.idx_vars:
            s.append("cdef int %s\n" % l.label)
        return s + ['\n']

    def _get_routine_ending(self, routine):
        return ["#end\n"]

    def _indent_code(self, codelines):
        p = CythonCodePrinter()
        return p.indent_code(codelines)

    def dump_pyx(self, routines, f, prefix, header=True, empty=True):
        self.dump_code(routines, f, prefix, header, empty)

    dump_pyx.extension = code_extension
    dump_pyx.__doc__ = CodeGen.dump_code.__doc__

    # This list of dump functions is used by CodeGen.write to know which dump
    # functions it has to call.
    dump_fns = [dump_pyx]


class NumpyCodeGen(LBMCodeGen):
    """Generator for Cython code.

    The .write() method inherited from CodeGen will output a code file <prefix>.pyx.

    """

    code_extension = "py"

    def code_generator(self, expr, assign_to=None, **settings):
        return numpy_code(expr, assign_to, **settings)

    def _get_header(self):
        code_lines = ["import numpy as np\n",
                     ]
        return code_lines + ["\n\n"]

    def _preprocessor_statements(self, prefix):
        return []

    def _get_routine_opening(self, routine):
        """Returns the opening statements of the routine."""
        code_list = []
        code_list.append("def ")

        # Inputs
        args = []
        for i, arg in enumerate(routine.arguments):
            if isinstance(arg, OutputArgument):
                raise CodeGenError("Numpy: invalid argument of type %s" %
                                   str(type(arg)))

            if isinstance(arg, (InputArgument, InOutArgument)):
                name = self._get_symbol(arg.name)
                args.append(name)
        args = ", ".join(args)
        code_list.append("%s(%s):\n" % (routine.name, args))
        code_list = [ "".join(code_list) ]

        return code_list

    def _declare_arguments(self, routine):
        return []

    def _declare_globals(self, routine):
        return []

    def _declare_locals(self, routine):
        return []

    def _get_routine_ending(self, routine):
        return ["#end\n"]

    def _indent_code(self, codelines):
        p = NumpyCodePrinter()
        return p.indent_code(codelines)

    def dump_py(self, routines, f, prefix, header=True, empty=True):
        self.dump_code(routines, f, prefix, header, empty)

    dump_py.extension = code_extension
    dump_py.__doc__ = CodeGen.dump_code.__doc__

    # This list of dump functions is used by CodeGen.write to know which dump
    # functions it has to call.
    dump_fns = [dump_py]

class LoopyCodeGen(LBMCodeGen):
    """Generator for Cython code.

    The .write() method inherited from CodeGen will output a code file <prefix>.pyx.

    """

    code_extension = "py"

    _default_settings = {"prefetch": None}

    def code_generator(self, expr, assign_to=None, **settings):
        return loopy_code(expr, assign_to, **settings)

    def _preprocessor_statements(self, prefix):
        return []

    def _get_header(self):
        code_lines = ["import loopy as lp\n",
                      "import numpy as np\n"
                     ]
        return code_lines + ["\n\n"]

    def _get_routine_opening(self, routine):
        """Returns the opening statements of the routine."""
        code_list = []
        code_list.append("%s = lp.make_kernel("%routine.name)
        name = []
        bounds = []
        for i in routine.idx_vars:
            if isinstance(i, Idx):
                name.append("%s_"%i.label)
                bounds.append("0<={ilabel}_<{upper}".format(ilabel=i.label, upper=i.upper-i.lower))

        if len(name) > 0:
            code_list.append('"{[%s]:%s}",'%(",".join(name), " and ".join(bounds)))
        code_list.append('"""  # noqa (silences flake8 line length warning)\n')
        code_list = [ "\n".join(code_list) ]
        return code_list

    def _declare_arguments(self, routine):
        return []

    def _declare_globals(self, routine):
        return []

    def _declare_locals(self, routine):
        return []

    def _get_routine_ending(self, routine):
        code_list = []
        code_list.append('""",')

        # Inputs
        args = []
        dtypes = []
        for i, arg in enumerate(routine.arguments):
            if isinstance(arg, OutputArgument):
                raise CodeGenError("Loopy: invalid argument of type %s" %
                                   str(type(arg)))

            if isinstance(arg, (InputArgument, InOutArgument)):
                name = self._get_symbol(arg.name)
                if arg.dimensions:
                    dims = ["{}".format(d[1]-d[0]+1) for d in arg.dimensions]
                    dtype = arg.get_datatype('PYTHON')
                    if dtype == 'int':
                        dtype = 'np.int32'
                    args.append('lp.GlobalArg("{name}", dtype={dtype}, shape="{shape}")'.format(name=name, dtype=dtype, shape=", ".join(dims)))
                else:
                    args.append('lp.ValueArg("{name}", dtype={dtype})'.format(name=name, dtype=arg.get_datatype('PYTHON')))
        for i, arg in enumerate(routine.local_vars):
            if isinstance(arg, Symbol):
                args.append('lp.TemporaryVariable("{name}", dtype=float)'.format(name=self._get_symbol(arg)))    
            else:
                dims = [d for d in arg.shape if d!=1]
                args.append('lp.TemporaryVariable("{name}", dtype=float, shape="{shape}")'.format(name=self._get_symbol(arg), shape=','.join("%s"%s for s in dims)))

        code_list.append('[')
        args = ",\n".join(args)
        code_list.append(args)
        code_list.append('])#endArg\n')

        # add type
        dim = len(routine.idx_vars)
        if dim == 1:
            block_size = [256]
        if dim == 2:
            block_size = [16, 16]
        if dim == 3:
            block_size = [4, 4, 4]

        for i, idx in enumerate(routine.idx_vars[-1::-1]):
            code_list.append('{name} = lp.split_iname({name}, "{label}", {block}, outer_tag="g.{ilabel}", inner_tag="l.{ilabel}")'.format(name = routine.name,
            label = "%s_"%idx.label, ilabel=i, block=block_size[i]))
        code_list.append('{name} = lp.expand_subst({name})\n'.format(name=routine.name))
        code_list.append('{name} = lp.set_options({name}, no_numpy = True)\n'.format(name=routine.name))

        prefetch = routine.settings.get("prefetch", None)

        if prefetch:
            for var in prefetch:
                indices = []
                for idx in routine.idx_vars:
                    indices.append("%s__inner"%idx.label)
                # for i in range(var.rank):
                #     if isinstance(var.indices[i], Idx):
                #         indices.append("%s__inner"%var.indices[i].label)
                code_list.append('{name} = lp.add_prefetch({name}, "{var}", "{label}", fetch_bounding_box=True)\n'.format(name=routine.name, var=var.base.label, label=",".join(indices)))
#            print("PREFETCH")
        #     label = []
        #     for var in self._settings["prefetch"]:
        #         indices = []
        #         for i in var.indices:
        #             if isinstance(i, Idx):
        #                 indices.append("%s__inner"%i.label)
        #         code_list.append('{name} = lp.add_prefetch({name}, "{var}", "{label}", fetch_bounding_box=True)'.format(name=routine.name, var=var.label, label=",".join(indices)))
        #print(LoopyCodePrinter()._sort_optimized(routine.local_vars, routine.instructions))
# one_time_step = lp.split_iname(one_time_step, "ii", 16, outer_tag="g.1", inner_tag="l.1")
# one_time_step = lp.split_iname(one_time_step, "jj", 16, outer_tag="g.0", inner_tag="l.0")
# one_time_step = lp.expand_subst(one_time_step)
# one_time_step = lp.add_prefetch(one_time_step, "f", "ii_inner,jj_inner", fetch_bounding_box=True)
        code_list = [ "\n".join(code_list) ]
        return code_list

    def _indent_code(self, codelines):
        p = LoopyCodePrinter()
        return p.indent_code(codelines)

    def dump_py(self, routines, f, prefix, header=True, empty=True):
        self.dump_code(routines, f, prefix, header, empty)

    dump_py.extension = code_extension
    dump_py.__doc__ = CodeGen.dump_code.__doc__

    # This list of dump functions is used by CodeGen.write to know which dump
    # functions it has to call.
    dump_fns = [dump_py]

def get_code_generator(language, project):
    CodeGenClass = {"NUMPY" : NumpyCodeGen,
                    "CYTHON": CythonCodeGen,
                    "LOOPY": LoopyCodeGen}.get(language.upper())
    if CodeGenClass is None:
        raise ValueError("Language '%s' is not supported." % language)
    return CodeGenClass(project)

def codegen(name_expr, language, prefix=None, project="project",
            to_files=False, header=True, empty=True, argument_sequence=None,
            global_vars=None, settings={}):
    # Initialize the code generator.
    code_gen = get_code_generator(language, project)

    if isinstance(name_expr[0], string_types):
        # single tuple is given, turn it into a singleton list with a tuple.
        name_expr = [name_expr]

    if prefix is None:
        prefix = name_expr[0][0]

    # Construct Routines appropriate for this code_gen from (name, expr) pairs.
    routines = []
    for name, expr in name_expr:
        routines.append(code_gen.routine(name, expr, argument_sequence,
                                         global_vars, settings))

    # Write the code.
    return code_gen.write(routines, prefix, to_files, header, empty)

def make_routine(name_expr, argument_sequence=None, local_vars=None, settings={}):
    if isinstance(name_expr[0], string_types):
        # single tuple is given, turn it into a singleton list with a tuple.
        name_expr = [name_expr]

    routines = []
    for name, expr in name_expr:
        routines.append(LBMCodeGen().routine(name, expr, argument_sequence,
                                             local_vars, settings))

    return routines



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Symbolic computation of equivalent equations
"""

#pylint: disable=invalid-name
import sympy as sp
from ..symbolic import alltogether

class EquivalentEquation:
    def __init__(self, scheme):
        # TODO: add source terms

        t, x, y, z = sp.symbols("t x y z", type='real')
        consm = list(scheme.consm.keys())
        nconsm = len(scheme.consm)

        self.consm = sp.Matrix(consm)
        self.dim = scheme.dim

        space = [x, y, z]

        LA = scheme.symb_la
        if not LA:
            LA = scheme.la

        func = []
        for i in range(nconsm):
            func.append(sp.Function('f{}'.format(i))(t, x, y, z)) #pylint: disable=not-callable
        func = sp.Matrix(func)

        sublist = [(i, j) for i, j in zip(consm, func)]
        sublist_inv = [(j, i) for i, j in zip(consm, func)]

        eq_func = sp.Matrix(scheme.EQ[nconsm:]).subs(sublist)
        s = sp.Matrix(scheme.s[nconsm:])

        all_vel = scheme.stencil.get_all_velocities()

        Lambda = []
        for i in range(all_vel.shape[1]):
            vd = LA*sp.diag(*all_vel[:, i])
            Lambda.append(scheme.M*vd*scheme.invM)

        phi1 = sp.zeros(s.shape[0], 1) #pylint: disable=unsubscriptable-object
        sigma = sp.diag(*s).inv() - sp.eye(len(s))/2
        gamma_1 = sp.zeros(nconsm, 1)

        self.coeff_order1 = []
        for dim, lambda_ in enumerate(Lambda):
            A, B = sp.Matrix([lambda_[:nconsm, :nconsm]]), sp.Matrix([lambda_[:nconsm, nconsm:]])
            C, D = sp.Matrix([lambda_[nconsm:, :nconsm]]), sp.Matrix([lambda_[nconsm:, nconsm:]])

            self.coeff_order1.append(A*func + B*eq_func)
            alltogether(self.coeff_order1[-1], nsimplify=True)
            for i in range(nconsm):
                gamma_1[i] += sp.Derivative(self.coeff_order1[-1][i], space[dim])

            dummy = -C*func - D*eq_func
            alltogether(dummy, nsimplify=True)
            for i in range(dummy.shape[0]):
                phi1[i] += sp.Derivative(dummy[i], space[dim])

        self.coeff_order2 = [[sp.zeros(nconsm) for _ in range(scheme.dim)] for _ in range(scheme.dim)]

        for dim, lambda_ in enumerate(Lambda):
            A, B = sp.Matrix([lambda_[:nconsm, :nconsm]]), sp.Matrix([lambda_[:nconsm, nconsm:]])

            meq = sp.Matrix(scheme.EQ[nconsm:])
            jac = meq.jacobian(consm)
            jac = jac.subs(sublist)
            delta1 = jac*gamma_1

            phi1_ = phi1 + delta1
            sphi1 = B*sigma*phi1_
            sphi1 = sphi1.doit()
            alltogether(sphi1, nsimplify=True)

            for i in range(scheme.dim):
                for jc in range(nconsm):
                    for ic in range(nconsm):
                        self.coeff_order2[dim][i][ic, jc] += sphi1[ic].expand().coeff(sp.Derivative(func[jc], space[i]))

        for ic, c in enumerate(self.coeff_order1):
            self.coeff_order1[ic] = c.subs(sublist_inv).doit()

        for ic, c in enumerate(self.coeff_order2):
            for jc, cc in enumerate(c):
                self.coeff_order2[ic][jc] = cc.subs(sublist_inv).doit()

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('equivalent_equation.tpl')
        t, x, y, z, U, Fx, Fy, Fz, Delta = sp.symbols('t, x, y, z, U, Fx, Fy, Fz, Delta_t')
        Bxx, Bxy, Bxz = sp.symbols('Bxx, Bxy, Bxz')
        Byx, Byy, Byz = sp.symbols('Byx, Byy, Byz')
        Bzx, Bzy, Bzz = sp.symbols('Bzx, Bzy, Bzz')

        phys_equation = sp.Derivative(U, t) + sp.Derivative(Fx, x)
        if self.dim > 1:
            phys_equation += sp.Derivative(Fy, y)
        if self.dim == 3:
            phys_equation += sp.Derivative(Fz, z)

        order2 = []
        space = [x, y, z]
        B = [[Bxx, Bxy, Bxz],
             [Byx, Byy, Byz],
             [Bzx, Bzy, Bzz],
            ]

        phys_equation_rhs = 0
        for i in range(self.dim):
            for j in range(self.dim):
                order2.append(sp.pretty(sp.Eq(B[i][j], -Delta*self.coeff_order2[i][j], evaluate=False)))
                phys_equation_rhs += sp.Derivative(B[i][j]*sp.Derivative(U, space[j]), space[i])
        return template.render(header=header_string('Equivalent Equations'),
                               dim=self.dim,
                               phys_equation=sp.pretty(sp.Eq(phys_equation, phys_equation_rhs)),
                               conserved_moments=sp.pretty(sp.Eq(U, self.consm, evaluate=False)),
                               order1=[sp.pretty(sp.Eq(F, coeff, evaluate=False)) for F, coeff in zip([Fx, Fy, Fz][:self.dim], self.coeff_order1)],
                               order2=order2
                              )

    def __repr__(self):
        return self.__str__()



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

from .base import BaseAlgorithm
from .pull import PullAlgorithm


# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause
queue = None #pylint: disable=invalid-name

def set_queue(backend):
    global queue #pylint: disable=invalid-name
    if backend.upper() == "LOOPY":
        try:
            import pyopencl as cl
        except ImportError:
            raise ImportError('Please install loo.py')
        ctx = cl.create_some_context()
        queue = cl.CommandQueue(ctx)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
HDF5 module
"""
import os
import logging
from six.moves import range
import numpy as np
import h5py
import mpi4py.MPI as mpi

log = logging.getLogger(__name__) #pylint: disable=invalid-name

class H5File:
    """
    class to manage hfd5 and xdmf file.
    """
    def __init__(self, mpi_topo, filename, path='', timestep=None, init_xdmf=False):
        if timestep:
            prefix = '_{}'.format(timestep)
        else:
            prefix = ''
        self.path = path
        name, ext = os.path.splitext(filename)
        self.filename = name + prefix
        self.h5filename = name + prefix + '.h5'

        self.origin = None
        self.dx = None
        self.dim = None
        self.n = None
        self.region = None
        self.global_size = None
        self.xdmf_file = None

        if mpi.COMM_WORLD.Get_rank() == 0:
            if not os.path.exists(path):
                os.mkdir(path)

            self.h5file = h5py.File(path + '/' + self.h5filename, "w")

        # All the processes wait for the creation of the output directory
        mpi.COMM_WORLD.Barrier()

        self.mpi_topo = mpi_topo
        self.scalars = {}
        self.vectors = {}
        self._init_grid = True
        self._init_xdmf = init_xdmf

    def set_grid(self, x, y=None, z=None):
        """
        create the hdf5 coordinate.

        Parameters
        ----------

        x : ndarray
            x-coordinate

        y : ndarray
            y-coordinate
            default is None

        z : ndarray
            z-coordinate
            default is None

        """
        if not self._init_grid:
            log.warning("h5 grid redefined.")

        self.origin = [x[0]]
        self.dx = [x[1] - x[0]]
        coords = [x]
        if y is not None:
            self.origin.append(y[0])
            self.dx.append(y[1] - y[0])
            coords.append(y)
        if z is not None:
            self.origin.append(z[0])
            self.dx.append(z[1] - z[0])
            coords.append(z)

        self.dim = len(self.origin)
        self.n = [0]*self.dim
        self.region = []
        self.global_size = []

        # get the region own by each processes
        # the global size
        # and the global coords
        for i in range(self.dim):
            sub = [False]*self.dim
            sub[i] = True
            comm = self.mpi_topo.cartcomm.Sub(sub)
            self.n[i] = comm.allreduce(coords[i].size, op=mpi.SUM)
            self.region.append(comm.gather(coords[i].size, root=0))
            self.global_size.append(np.sum(np.asarray(self.region[i])))
            coords[i] = np.asarray(comm.gather(coords[i], root=0))
            comm.Free()

        if mpi.COMM_WORLD.Get_rank() == 0:
            for i in range(len(self.region)):
                self.region[i].insert(0, 0)
                for j in range(1, len(self.region[i])):
                    self.region[i][j] += self.region[i][j-1]
            #self.region = np.asarray(self.region, dtype=np.int)
            #self.region = np.concatenate((np.zeros((self.dim, 1), dtype=np.int), np.cumsum(self.region, axis=1)), axis=1)
            #print(self.region)
            for i in range(self.dim):
                dset = self.h5file.create_dataset("x_{}".format(i), [self.global_size[i]], dtype=np.double)
                dset[:] = np.concatenate(coords[i])

    def _get_slice(self, rank):
        """
        get the part own by the processor rank and
        the size of the data in each direction
        """
        ind = []
        buffer_size = []
        mpi_coords = self.mpi_topo.cartcomm.Get_coords(rank)
        # print('coords', rank, mpi_coords)
        for i in range(self.dim):
            ind.append(slice(self.region[i][mpi_coords[i]], self.region[i][mpi_coords[i]+1]))
            buffer_size.append(self.region[i][mpi_coords[i]+1] - self.region[i][mpi_coords[i]])
        return ind[::-1], buffer_size[::-1]

    def _set_dset(self, dset, comm, data, index=0, with_index=False):
        """
        Merge data from multiple sub domains into a dataset.

        Parameters
        ----------

        dset : dataset
            hdf5 dataset where to store the data

        comm : comm
            mpi communicator

        data : array
            data on the sub-domain

        index : int
            use to store an index of a vector

        with_index : bool
            if we store a vector component

        """
        ind, buffer_size = self._get_slice(0)
        ind = tuple(ind)
        if with_index:
            ind = ind + (index,)
        dset[ind] = data.T

        for i in range(1, comm.Get_size()):
            ind, buffer_size = self._get_slice(i)
            #print(i, ind, buffer_size)
            ind = tuple(ind)
            if with_index:
                ind = ind + (index,)
            rcv_buffer = np.empty(buffer_size)
            comm.Recv([rcv_buffer, mpi.DOUBLE], source=i, tag=index)
            dset[ind] = rcv_buffer

    def add_scalar(self, name, f, *fargs):
        """
        store a scalar field.

        Parameters
        ----------

        name: string
            the name of the dataset entry

        f : array or function
            data to store

        fargs: tuple
            arguments of f if f is a function

        """
        if isinstance(f, np.ndarray):
            data = f
        else:
            data = f(*fargs)

        comm = self.mpi_topo.cartcomm
        if comm.Get_rank() == 0:
            dset = self.h5file.create_dataset(name, self.global_size[::-1], dtype=np.double)
            self._set_dset(dset, comm, data)
            self.scalars[name] = self.h5filename + ":/" + name
        else:
            comm.Send([np.ascontiguousarray(data.T, dtype=np.double), mpi.DOUBLE], dest=0, tag=0)

    def add_vector(self, name, f, *fargs):
        """
        store a vector field.

        Parameters
        ----------

        name: string
            the name of the dataset entry

        f : array or function
            data to store

        fargs: tuple
            arguments of f if f is a function

        """
        if isinstance(f, list):
            datas = f
        else:
            datas = f(*fargs)

        comm = self.mpi_topo.cartcomm
        if comm.Get_rank() == 0:
            dset = self.h5file.create_dataset(name, self.global_size[::-1] + [3], dtype=np.double)
            for i, data in enumerate(datas):
                self._set_dset(dset, comm, data, i, with_index=True)
            self.vectors[name] = self.h5filename + ":/" + name
        else:
            for i, data in enumerate(datas):
                comm.Send([np.ascontiguousarray(data.T), mpi.DOUBLE], dest=0, tag=i)

    def save(self):
        """
        save the hdf5 and the xdmf files.
        """
        comm = self.mpi_topo.cartcomm
        if comm.Get_rank() == 0:
            self.h5file.close()
            self.xdmf_file = open(self.path + '/' + self.filename + '.xdmf', "w")
            self.xdmf_file.write("""<?xml version="1.0" ?>
<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
<Xdmf>
 <Domain>           
            """)
            if self.dim == 2:
                self.xdmf_file.write("""
                <Grid Name="Structured Grid" GridType="Uniform">
                    <Topology TopologyType="2DRectMesh" NumberOfElements="{0}"/>
                    <Geometry GeometryType="VXVY">
                """.format(' '.join(map(str, self.global_size[::-1]))))
            else:
                self.xdmf_file.write("""
                <Grid Name="Structured Grid" GridType="Uniform">
                    <Topology TopologyType="3DRectMesh" NumberOfElements="{0}"/>
                    <Geometry GeometryType="VXVYVZ">
                """.format(' '.join(map(str, self.global_size[::-1]))))
            for i in range(self.dim-1, -1, -1):
                self.xdmf_file.write("""
                <DataItem Format="HDF" Dimensions="{0}">
                    {1}:/x_{2} 
                </DataItem>
                """.format(self.global_size[i], self.filename + '.h5', i))

            self.xdmf_file.write("</Geometry>\n")

            for k, v in self.scalars.items():
                self.xdmf_file.write("""
                <Attribute Name="{0}" AttributeType="Scalar" Center="Node">
                <DataItem Format="HDF" Dimensions="{1}">
                {2} 
                </DataItem>
                </Attribute>
                """.format(k, ' '.join(map(str, self.global_size[::-1])), v))

            for k, v in self.vectors.items():
                self.xdmf_file.write("""
                <Attribute Name="{0}" AttributeType="Vector" Center="Node">
                <DataItem Format="HDF" Dimensions="{1} {2}">
                {3} 
                </DataItem>
                </Attribute>
                """.format(k, ' '.join(map(str, self.global_size[::-1])), self.dim, v))

            self.xdmf_file.write("</Grid>\n</Domain>\n</Xdmf>\n")
            self.xdmf_file.close()



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
module monitoring
usage: python filename.py --monitoring
"""

# import time
import inspect
from functools import wraps
import atexit
import numpy as np
import mpi4py.MPI as mpi

from .options import options


class PerfMonitor:
    def __init__(self):
        self.count = 0
        self.total_time = []
        self.self_time = []


class Node:
    def __init__(self, parent=None, name='root'):
        self.parent = parent
        self.name = name
        self.sons = []
        self.time = 0

    def add_node(self, name):
        self.sons.append(Node(self, name))
        return self.sons[-1]

    def del_node(self):
        self.sons.clear()
        return self.parent

    def add_time(self, t):
        self.parent.time += t


class Monitoring:
    def __init__(self):
        self.func = {}
        self.size = 0
        self.tree = Node()

    @staticmethod
    def information(f):
        mod_name = inspect.getmodule(f).__name__
        try:
            f_name = f.__qualname__
            return (mod_name, f_name)
        except:
            pass

    def set_size(self, size):
        self.size = size

    def register(self, f):
        info = self.information(f)
        if not self.func.get(info, None):
            self.func[info] = PerfMonitor()

    def start_timing(self, f):
        info = self.information(f)
        self.tree = self.tree.add_node(info)
        self.func[info].total_time.append(mpi.Wtime())
        self.func[info].self_time.append(0)

    def stop_timing(self, f):
        info = self.information(f)
        t = mpi.Wtime()
        self.func[info].total_time[-1] = t - self.func[info].total_time[-1]
        self.tree.add_time(self.func[info].total_time[-1])
        self.func[info].self_time[-1] = self.func[info].total_time[-1] \
            - self.tree.time
        self.tree = self.tree.del_node()

    def __str__(self):
        if mpi.COMM_WORLD.rank == 0:
            titles = [
                '%', 'module name', 'function name',
                'ncall', 'total time', 'self time', 'MLUPS'
            ]
            row_format = "{:>6}{:>25}{:>30}{:>8}{:>15}{:>15}{:>8}"
            print('\n', row_format.format(*titles), '\n')
            row_format = "{:6.1f}{:>25}{:>30}{:8}{:15.9f}{:15.9f}{:8.2f}"
            names = [[k[0], k[1]] for k in self.func.keys()]
            data = [
                [
                    len(v.total_time),
                    np.sum(v.total_time),
                    np.sum(v.self_time)
                ] for v in self.func.values()
            ]
            ind = np.argsort(np.asarray(data)[:, 1])
            for i in ind[::-1]:
                print(row_format.format(data[i][1]/data[ind[-1]][1]*100,
                                        *names[i],
                                        *data[i],
                                        data[i][0]*self.size/data[i][1]/1e6))

Monitor = Monitoring()  # pylint: disable=invalid-name

if options().monitoring:
    atexit.register(Monitor.__str__)


def monitor(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        Monitor.register(f)
        Monitor.start_timing(f)
        output = f(*args, **kwds)
        Monitor.stop_timing(f)
        return output
    return wrapper



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Ellipsoid element
"""

# pylint: disable=invalid-name

import logging
# from textwrap import dedent
import numpy as np

from .base import Element
from .utils import distance_ellipsoid

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Ellipsoid(Element):
    """
    Class Ellipsoid

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        a vector
    v2 : list
        a vector
    v3 : list
        a vector (v1, v2, and v3 have to be orthogonal)
    label : list
        one integer (default [0])
    isfluid : boolean
        - True if the ellipsoid is added
        - False if the ellipsoid is deleted

    Attributes
    ----------
    number_of_bounds : int
        1
    dim: int
        3
    center : ndarray
        the coordinates of the center of the sphere
    v1 : ndarray
        the coordinates of the first vector
    v2 : ndarray
        the coordinates of the second vector
    v3 : ndarray
        the coordinates of the third vector
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the ellipsoid is added
        and False if the ellipsoid is deleted

    Examples
    --------

    the ellipsoid centered in (0, 0, 0)
    with v1=[3,0,0], v2=[0,2,0], and v3=[0,0,1]

    >>> center = [0., 0., 0.]
    >>> v1, v2, v3 = [3,0,0], [0,2,0], [0,0,1]
    >>> Ellipsoid(center, v1, v2, v3)
    +-----------+
    | Ellipsoid |
    +-----------+
        - dimension: 3
        - center: [0. 0. 0.]
        - v1: [3 0 0]
        - v2: [0 2 0]
        - v3: [0 0 1]
        - label: [0]
        - type: solid

    """
    def __init__(self, center, v1, v2, v3, label=0, isfluid=False):
        self.number_of_bounds = 1  # number of edges
        self.dim = 3
        self.center = np.asarray(center)
        p12 = abs(v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2])
        p23 = abs(v2[0]*v3[0] + v2[1]*v3[1] + v2[2]*v3[2])
        p31 = abs(v3[0]*v1[0] + v3[1]*v1[1] + v3[2]*v1[2])
        if max(p12, p23, p31) > 1.e-14:
            log.error('The vectors of the ellipsoid are not orthogonal')
        else:
            self.v1 = np.asarray(v1)
            self.v2 = np.asarray(v2)
            self.v3 = np.asarray(v3)
        super(Ellipsoid, self).__init__(label, isfluid)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the ellipsoid.
        """
        r = max(np.linalg.norm(self.v1),
                np.linalg.norm(self.v2),
                np.linalg.norm(self.v3))
        return self.center - r, self.center + r

    # pylint: disable=too-many-locals
    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the ellipsoid.

        Notes
        -----

        the edge of the ellipsoid is considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the ellipsoid, False otherwise)

        """
        x, y, z = grid

        X = x - self.center[0]
        Y = y - self.center[1]
        Z = z - self.center[2]
        v12 = np.cross(self.v1, self.v2)
        v23 = np.cross(self.v2, self.v3)
        v31 = np.cross(self.v3, self.v1)
        d = np.inner(self.v1, v23)**2
        # equation of the ellipsoid:
        # cxx XX + cyy YY + czz ZZ + cxy XY + cyz YZ + czx ZX = d
        cxx = v12[0]**2 + v23[0]**2 + v31[0]**2
        cyy = v12[1]**2 + v23[1]**2 + v31[1]**2
        czz = v12[2]**2 + v23[2]**2 + v31[2]**2
        cxy = 2 * (v12[0]*v12[1] + v23[0]*v23[1] + v31[0]*v31[1])
        cyz = 2 * (v12[1]*v12[2] + v23[1]*v23[2] + v31[1]*v31[2])
        czx = 2 * (v12[2]*v12[0] + v23[2]*v23[0] + v31[2]*v31[0])
        return cxx*X**2 + cyy*Y**2 + czz*Z**2 + \
            cxy*X*Y + cyz*Y*Z + czx*Z*X <= d

    def distance(self, grid, v, dmax=None):
        """
        Compute the distance in the v direction between
        the ellipsoid and the points defined by (x, y, z).

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max

        Returns
        -------

        ndarray
            array of distances

        """
        x, y, z = grid
        return distance_ellipsoid(
            x, y, z, v,
            self.center, self.v1, self.v2, self.v3,
            dmax, self.label
        )

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('ellipsoid.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(
            header=header_string(self.__class__.__name__),
            elem=self, type=elem_type
        )

    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(3), alpha=1.
                  ):
        v1 = scale*self.v1
        v2 = scale*self.v2
        v3 = scale*self.v3
        viewer.ellipse_3d(self.center*scale, v1, v2, v3, color, alpha=alpha)
        if viewlabel:
            x, y, z = self.center[0], self.center[1], self.center[2]
            viewer.text(str(self.label[0]), [x, y, z])



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
A flexible Python package for lattice Boltzmann method.
"""
import logging
from colorlog import ColoredFormatter
import mpi4py.MPI as mpi

from colorama import init
init()

from .domain import Domain
from .stencil import Stencil
from .simulation import Simulation
from . import boundary as bc
from .scheme import Scheme
from .elements import * #pylint: disable=wildcard-import
from .geometry import Geometry
from . import viewer
from .hdf5 import H5File
from .options import options
from . import monitoring
from .analysis import EquivalentEquation, Stability

from .version import version as __version__

numeric_level = getattr(logging, options().loglevel, None) #pylint: disable=invalid-name

#pylint: disable=bad-continuation
formatter = ColoredFormatter( #pylint: disable=invalid-name
        "%(log_color)s[{}] %(levelname)-8s %(name)s in function %(funcName)s line %(lineno)s\n%(reset)s%(message)s".format(mpi.COMM_WORLD.Get_rank()), #pylint: disable=too-many-format-args
        datefmt=None,
        reset=True,
        log_colors={'DEBUG':    'green',
                    'INFO':     'cyan',
                    'WARNING':  'blue',
                    'ERROR':    'red',
                    'CRITICAL': 'red,bg_white',
                   },
        style='%'
        )

logger = logging.getLogger(__name__) #pylint: disable=invalid-name
logger.setLevel(level=numeric_level)

console = logging.StreamHandler() #pylint: disable=invalid-name
console.setFormatter(formatter)
logger.addHandler(console)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Description of a LBM scheme
"""
import sys
import logging
from textwrap import dedent

import sympy as sp
from sympy.parsing.sympy_parser import parse_expr
from sympy import symbols, Eq

from .stencil import Stencil
from .validator import validate
from .symbolic import rel_ux, rel_uy, rel_uz, alltogether
#pylint: disable=too-many-lines

log = logging.getLogger(__name__) #pylint: disable=invalid-name

def allfactor(M):
    """
    Factorize all the elements of sympy matrix M

    Parameters
    ----------

    M : sympy matrix
       matrix to factorize

    """
    for i in range(M.shape[0]):
        for j in range(M.shape[1]):
            M[i, j] = M[i, j].factor()

def param_to_tuple(param):
    """
    Convert param dictionary to a list of keys and a list of values.

    Parameters
    ----------

    param : dictionary
        parameters

    Returns
    -------

    list
        the keys of param
    list
        the values of param

    """

    if param is not None:
        keys, values = list(param.keys()), list(param.values())
    else:
        keys, values = (), ()
    return keys, values

#pylint: disable=invalid-name, too-many-locals, too-many-branches, too-many-statements
class Scheme:
    """
    Create the class with all the needed informations for each elementary scheme.

    Parameters
    ----------

    dico : a dictionary that contains the following `key:value`
      - dim : spatial dimension (optional if the `box` is given)
      - scheme_velocity : the value of the ratio space step over time step
        (la = dx / dt)
      - schemes : a list of dictionaries, one for each scheme

    Notes
    -----

    Each dictionary of the list `schemes` should contains the following `key:value`

    - velocities : list of the velocities number
    - conserved moments : list of the moments conserved by each scheme
    - polynomials : list of the polynomial functions that define the moments
    - equilibrium : list of the values that define the equilibrium
    - relaxation_parameters : list of the value of the relaxation parameters
    - source_terms : dictionary do define the source terms (optional, see examples)
    - init : dictionary to define the initial conditions (see examples)

    If the stencil has already been computed, it can be pass in argument.

    Attributes
    ----------

    dim : int
      spatial dimension
    dx : double
      space step
    dt : double
      time step
    la : double
      scheme velocity, ratio dx/dt
    nschemes : int
      number of elementary schemes
    stencil : object of class :py:class:`Stencil <pylbm.stencil.Stencil>`
      a stencil of velocities
    P : list of sympy matrix
      list of polynomials that define the moments
    EQ : list of sympy matrix
      list of the equilibrium functions
    s  : list of list of doubles
      relaxation parameters
      (exemple: s[k][l] is the parameter associated to the lth moment in the kth scheme)
    M : sympy matrix
      the symbolic matrix of the moments
    Mnum : numpy array
      the numeric matrix of the moments (m = Mnum F)
    invM : sympy matrix
      the symbolic inverse matrix
    invMnum : numpy array
      the numeric inverse matrix (F = invMnum m)

    Examples
    --------

    see demo/examples/scheme/

    """
    def __init__(self,
                 dico,
                 check_inverse=False,
                 need_validation=True):

        if need_validation:
            validate(dico, __class__.__name__)

        self.stencil = Stencil(dico, need_validation=False)
        self.dim = self.stencil.dim

        # symbolic parameters
        self.param = dico.get('parameters', {})

        self.symb_la, self.la = self._get_var_symbolic(dico, 'scheme_velocity')
        if not self.la:
            log.error("The entry 'scheme_velocity' is wrong.")

        # set relative velocity
        self.rel_vel = dico.get('relative_velocity', None)

        # fix the variables of time and space
        self.symb_t, self.symb_coord = self._get_space_and_time_symbolic()

        self.nschemes = self.stencil.nstencils
        scheme = dico['schemes']

        self._check_entry_size(scheme, 'polynomials')
        self._check_entry_size(scheme, 'relaxation_parameters')

        self.P = sp.Matrix([p for s in scheme for p in s['polynomials']])
        self.s = sp.Matrix([r for s in scheme for r in s['relaxation_parameters']])
        self._source_terms = [s.get('source_terms', None) for s in scheme]
        self.M, self.invM, self.Tu, self.Tmu = self._create_moments_matrices()
        self.EQ = self._get_equilibrium(scheme)

        self.s_no_swap = self.s.copy()
        self.EQ_no_swap = self.EQ.copy()
        self.M_no_swap = self.M.copy()
        self.invM_no_swap = self.invM.copy()

        self.consm = self._get_conserved_moments(scheme)
        # put conserved moments at the beginning of EQ and s
        # and permute M, invM, Tu, and Tmu accordingly
        self._permute_consm_in_front()
        if self.rel_vel is not None:
            self.Tu_no_swap = self.Tu.copy()

        self._check_inverse(self.M, self.invM, 'M')
        self._check_inverse_of_Tu()

        log.info(self.__str__())

    def _get_var_symbolic(self, dico, entry):
        value = dico.get(entry, None)
        symb, num = None, None
        if isinstance(value, (int, float)):
            num = value
        elif isinstance(value, sp.Symbol):
            symb = value
            num = float(value.subs(self.param.items()))
        return symb, num

    def _get_space_and_time_symbolic(self):
        symb_t = self.param.get('time', sp.Symbol('t'))

        symb_coord = [None, None, None]
        symb_coord[0] = self.param.get('space_x', sp.Symbol('X'))
        symb_coord[1] = self.param.get('space_y', sp.Symbol('Y'))
        symb_coord[2] = self.param.get('space_z', sp.Symbol('Z'))

        return symb_t, symb_coord

    def _get_equilibrium(self, scheme):
        eq = []
        for i, s in enumerate(scheme):
            feq = s.get('feq', None)
            meq = s.get('equilibrium', None)
            if feq and meq:
                log.error("Error in the creation of the scheme %d: you can have only 'feq' or 'equilibrium'", i)
                sys.exit()
            if meq:
                eq.append(meq)
            if feq:
                sli = slice(self.stencil.nv_ptr[i], self.stencil.nv_ptr[i+1])
                meq_tmp = self.M[sli, sli]*feq[0](self.stencil.get_all_velocities(i), *feq[1])
                meq_tmp.simplify()
                eq.append([e for e in meq_tmp])
        return sp.Matrix([e for sublist in eq for e in sublist])

    def _permute_consm_in_front(self):
        self.permutations = []
        for ic, c in enumerate(self.consm.values()):
            self.permutations.append([ic, c])

        for p in self.permutations:
            self.EQ.row_swap(p[0], p[1])
            self.s.row_swap(p[0], p[1])
            self.M.row_swap(p[0], p[1])
            self.invM.col_swap(p[0], p[1])
            self.Tu.row_swap(p[0], p[1])
            self.Tu.col_swap(p[0], p[1])
            self.Tmu.row_swap(p[0], p[1])
            self.Tmu.col_swap(p[0], p[1])

        for ic, c in enumerate(self.consm.keys()):
            self.consm[c] = ic

    def _check_entry_size(self, schemes, key):
        for i, s in enumerate(schemes):
            ls = len(s[key])
            nv = self.stencil.nv[i]
            if ls != nv:
                log.error(dedent("""\
                                the size of the entry for the key {0} in the scheme {1}
                                has not the same size of the stencil {1}: {2}, {3}""".format(key, i, ls, nv)))

    def __str__(self):
        from .utils import header_string
        from .jinja_env import env
        template = env.get_template('scheme.tpl')
        P = []
        EQ = []
        s = []
        header_scheme = []
        for k in range(self.nschemes):
            myslice = slice(self.stencil.nv_ptr[k], self.stencil.nv_ptr[k+1])
            header_scheme.append(header_string("Scheme %d"%k))
            P.append(sp.pretty(sp.Matrix(self.P[myslice])))
            EQ.append(sp.pretty(sp.Matrix(self.EQ_no_swap[myslice])))
            s.append(sp.pretty(sp.Matrix(self.s_no_swap[myslice])))

        if self.rel_vel:
            addons = {'rel_vel': self.rel_vel,
                      'Tu': sp.pretty(self.Tu_no_swap)
                     }
        else:
            addons = {}

        return template.render(header=header_string("Scheme information"),
                               scheme=self,
                               consm=sp.pretty(list(self.consm.keys())),
                               header_scheme=header_scheme,
                               P=P,
                               EQ=EQ,
                               s=s,
                               M=sp.pretty(self.M_no_swap),
                               invM=sp.pretty(self.invM_no_swap),
                               **addons
                              )

    def __repr__(self):
        return self.__str__()

    def _create_moments_matrices(self):
        """
        Create the moments matrices M and M^{-1} used to transform the repartition functions into the moments

        Three versions of these matrices are computed:

          - a sympy version M and invM for each scheme
          - a numerical version Mnum and invMnum for each scheme
          - a global numerical version MnumGlob and invMnumGlob for all the schemes
        """
        M_, invM_, Mu_, Tu_ = [], [], [], []
        u_tild = sp.Matrix([rel_ux, rel_uy, rel_uz])

        if self.symb_la is not None:
            LA = self.symb_la
        else:
            LA = self.la

        compt = 0
        for iv, v in enumerate(self.stencil.v):
            p = self.P[self.stencil.nv_ptr[iv] : self.stencil.nv_ptr[iv+1]]
            compt += 1
            lv = len(v)
            M_.append(sp.zeros(lv, lv))
            Mu_.append(sp.zeros(lv, lv))
            for i in range(lv):
                for j in range(lv):
                    sublist = [(str(self.symb_coord[d]), sp.Integer(v[j].v[d])*LA) for d in range(self.dim)]
                    M_[-1][i, j] = p[i].subs(sublist)

                    if self.rel_vel is not None:
                        sublist = [(str(self.symb_coord[d]), v[j].v[d]*LA - u_tild[d]) for d in range(self.dim)]
                        Mu_[-1][i, j] = p[i].subs(sublist)

            invM_.append(M_[-1].inv())
            Tu_.append(Mu_[-1]*invM_[-1])

        gshape = (self.stencil.nv_ptr[-1], self.stencil.nv_ptr[-1])
        Tu = sp.eye(gshape[0])
        M = sp.zeros(*gshape)
        invM = sp.zeros(*gshape)

        try:
            for k in range(self.nschemes):
                nvk = self.stencil.nv[k]
                for i in range(nvk):
                    for j in range(nvk):
                        index = self.stencil.nv_ptr[k] + i, self.stencil.nv_ptr[k] + j
                        M[index] = M_[k][i, j]
                        invM[index] = invM_[k][i, j]

                        if self.rel_vel is not None:
                            Tu[index] = Tu_[k][i, j]
        except TypeError:
            log.error("Unable to convert to float the expression %s or %s.\nCheck the 'parameters' entry.", M[k][i, j], invM[k][i, j]) #pylint: disable=undefined-loop-variable
            sys.exit()

        alltogether(Tu, nsimplify=True)
        alltogether(M, nsimplify=True)
        alltogether(invM, nsimplify=True)
        Tmu = Tu.subs(list(zip(u_tild, -u_tild)))
        return M, invM, Tu, Tmu

    def _check_inverse_of_Tu(self):
        # verification
        res = self.Tu*self.Tmu
        alltogether(res)
        gshape = self.stencil.nv_ptr[-1]
        test = res == sp.eye(gshape)
        if not test:
            log.warning("The property on the translation matrix is not verified\n T(u) * T(-u) is not identity !!!")

    def _check_inverse(self, M, invM, matrix_name):
        """
        Check if the product of a sympy matrix and its inverse
        is identity.

        Parameters
        ----------

        M : sympy.Matrix
            the matrix

        invM : sympy.Matrix
            the inverse of M

        matrix_name : string
            the name of the matrix

        FIXME: the name of the matrix should be obtain via M.

        """
        # verification
        gshape = self.stencil.nv_ptr[-1]
        dummy = M*invM
        alltogether(dummy)
        test = dummy == sp.eye(gshape)
        if not test:
            log.warning("Problem %s * inv%s is not identity !!!", matrix_name, matrix_name)

    def _get_conserved_moments(self, scheme):
        """
        return conserved moments and their indices in the scheme entry.

        Parameters
        ----------

        scheme : dictionnary
            description of the LBM schemes

        Returns
        -------

        dictionary
            the keys are the conserved moments and
            the values their indices in the LBM schemes.

        """
        from collections import OrderedDict
        consm_tmp = [s.get('conserved_moments', None) for s in scheme]
        consm = OrderedDict()

        for i in range(len(self.stencil.nv_ptr)-1):
            leq = self.EQ[self.stencil.nv_ptr[i]:self.stencil.nv_ptr[i+1]]
            cm_ieq = consm_tmp[i]
            if cm_ieq is not None:
                if isinstance(cm_ieq, (sp.Symbol, sp.IndexedBase)):
                    consm[cm_ieq] = self.stencil.nv_ptr[i] + leq.index(cm_ieq)
                else:
                    for c in cm_ieq:
                        consm[c] = self.stencil.nv_ptr[i] + leq.index(c)
        return consm

    def _get_indices_cons_noncons(self):
        """
        return the list of the conserved moments and the list of the non conserved moments

        Returns
        -------

        list
            the indices of the conserved moments
        list
            the indices of the non conserver moments

        """

        ns = self.stencil.nstencils # number of stencil
        nv = self.stencil.nv # number of velocities for each stencil
        l_cons = [[] for n in nv]
        l_noncons = [list(range(n)) for n in nv]
        for vk in list(self.consm.values()):
            l_cons[vk[0]].append(vk[1])
            l_noncons[vk[0]].remove(vk[1])
        for n in range(ns):
            l_cons[n].sort()
            l_noncons[n].sort()
        return l_cons, l_noncons

    def set_source_terms(self, scheme):
        """
        set the source terms functions for the conserved moments.

        Parameters
        ----------

        scheme : dictionnary
            description of the LBM schemes

        Returns
        -------

        dictionnary
            the keys are the indices of the conserved moments
            and the values must be a sympy expression or None

        """
        source_terms = []
        is_empty = True
        for ns, s in enumerate(scheme):
            source_terms.append([None]*self.stencil.nv[ns]) # by default no source term
            source_scheme = s.get('source_terms', None)
            if source_scheme is not None:
                for k, v in s['source_terms'].items():
                    try:
                        if isinstance(k, str):
                            indices = self.consm[parse_expr(k)]
                        elif isinstance(k, sp.Symbol):
                            indices = self.consm[k]
                        elif isinstance(k, int):
                            indices = (ns, k)
                        else:
                            raise ValueError
                    except ValueError:
                        log.error(dedent("""\
                                         Error in the creation of the scheme: wrong dictionnary
                                         the key `source_terms` should contain a dictionnary with
                                            key: the moment concerned
                                                 should be the name of the moment as a string or
                                                 a sympy Symbol or an integer
                                            value: the value of the source term
                                                 should be a float or a sympy expression
                                         """))
                        sys.exit()
                    source_terms[-1][indices[1]] = v
                    is_empty = False
        if is_empty:
            return None
        else:
            return source_terms



from setuptools import setup, find_packages

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Cython",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS"]


MAJOR = "0"
MINOR = "4"
PATCH = "0"
VERSION = "{0}.{1}.{2}".format(MAJOR, MINOR, PATCH)

def write_version_py(filename='pylbm/version.py'):
    a = open(filename, 'w')
    try:
        a.write("version = '{}'".format(VERSION))
    finally:
        a.close()

with open("README.rst", "r") as fh:
    long_description = fh.read()

write_version_py()

setup(
    name           = "pylbm",
    version        = VERSION,
    description    = "A flexible Python package for lattice Boltzmann method",
    long_description = long_description,
    author         = "Benjamin Graille, Loic Gouarin",
    author_email   = "benjamin.graille@math.u-psud.fr, loic.gouarin@polytechnique.edu",
    url            = "https://github.com/pylbm/pylbm",
    license        = "BSD",
    keywords       = "Lattice Boltzmann Methods",
    classifiers    = CLASSIFIERS,
    packages       = find_packages(exclude=['demo', 'doc', 'tests*']),
    package_data   = {'pylbm': ['templates/*']},
    include_package_data=True,
    install_requires=[
                        "numpy",
                        "matplotlib",
                        "sympy >=1.1.1,<1.2",
                        "cython",
                        "h5py",
                        "mpi4py",
                        "colorlog",
                        "colorama",
                        "cerberus",
                        "jinja2",
                      ],
)



# FIXME: make pylint happy !
#pylint: disable=all

import sys
import os
import shutil
import tempfile
from subprocess import STDOUT, CalledProcessError, check_output

from .codegen import get_code_generator

class CodeWrapError(Exception):
    pass


class CodeWrapper(object):
    """Base Class for code wrappers"""
    _module_name = "wrapped_module"
    _module_counter = 0

    @property
    def filename(self):
        return "%s_%s" % (self._module_name, CodeWrapper._module_counter)

    @property
    def module_name(self):
        return "%s_%s" % (self._module_name, CodeWrapper._module_counter)

    def __init__(self, generator, filepath=None, flags=[], verbose=False):
        """
        generator -- the code generator to use
        """
        self.generator = generator
        self.filepath = filepath
        self.flags = flags
        self.verbose = verbose

    def _generate_code(self, routines):
        self.generator.write(
            routines, self.filename, True, True, False)

    def wrap_code(self, routines):
        workdir = self.filepath or tempfile.mkdtemp("_sympy_compile")
        if not os.access(workdir, os.F_OK):
            os.mkdir(workdir)
        oldwork = os.getcwd()
        os.chdir(workdir)

        try:
            sys.path.append(workdir)
            self._prepare_files(routines)
            self._generate_code(routines)
            self._process_files(routines)
            mod = __import__(self.module_name)
        finally:
            sys.path.remove(workdir)
            CodeWrapper._module_counter += 1
            os.chdir(oldwork)
            if not self.filepath:
                try:
                    shutil.rmtree(workdir)
                except OSError:
                    # Could be some issues on Windows
                    pass

        return mod

    def _process_files(self, routine):
        command = self.command
        command.extend(self.flags)

        try:
            retoutput = check_output(command, stderr=STDOUT)
        except CalledProcessError as e:
            raise CodeWrapError(
                "Error while executing command: %s. Command output is:\n%s" % (
                    " ".join(command), e.output.decode()))
        if self.verbose:
            print(retoutput)


class CythonCodeWrapper(CodeWrapper):
    @property
    def command(self):
        bld = open(self.filename + '.pyxbld', "w")
        code = """
def make_ext(modname, pyxfilename):
    from distutils.extension import Extension

    return Extension(name = modname,
                     sources=[pyxfilename],
                     #extra_compile_args = ['-O3', '-fopenmp, '-w'],
                     #extra_link_args= ['-fopenmp'])
                     extra_compile_args = ['-O3', '-w']
                     #extra_compile_args = ['-O3', '-fopenmp', '-w'],
                     #extra_link_args= ['-fopenmp'])
                    )
                    """
        bld.write(code)
        bld.close()
        bld = open('build.py', 'w')
        code = """
import pyximport
pyximport.install(build_dir=\'.\', inplace=True)
import {module_name}
        """.format(module_name=self.filename)
        bld.write(code)
        bld.close()

        if self.verbose:
            print(open(self.filename + '.pyx').read())

        command = [sys.executable, 'build.py']

        return command

    def _prepare_files(self, routines):
        pass

class PythonCodeWrapper(CodeWrapper):
    @property
    def command(self):
        return []

    def _prepare_files(self, routines):
        pass

    def _process_files(self, routines):
        if self.verbose:
            print(open(self.filename + '.py').read())

def get_code_wrapper(backend):
    CodeWrapClass = {"NUMPY" : PythonCodeWrapper,
                     "CYTHON": CythonCodeWrapper,
                     "LOOPY": PythonCodeWrapper}.get(backend.upper())
    if CodeWrapClass is None:
        raise ValueError("Language '%s' is not supported." % backend)
    return CodeWrapClass

def autowrap(routines, backend='cython', tempdir=None, args=None, flags=[],
    verbose=False):

    code_generator = get_code_generator(backend, "project")
    CodeWrapperClass = get_code_wrapper(backend)
    code_wrapper = CodeWrapperClass(code_generator, tempdir, flags, verbose)

    return code_wrapper.wrap_code(routines)




# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

import sympy as sp
from sympy import Eq

def euler(lhs, rhs):
    """
    Explicit Euler in symbolic form.
    """
    dt = sp.Symbol('dt')
    return Eq(lhs, lhs + dt/2*rhs)